###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Base implementation for interaction with the iLO interface"""

# ---------Imports---------

import logging
import os
import struct
import time
from ctypes import byref, c_uint32, c_char_p, c_void_p, create_string_buffer

# ---------End of imports---------
# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------


class BlobReturnCodes(object):
    """Blob store return codes.

    SUCCESS           success

    """

    SUCCESS = 0
    if os.name != "nt":
        CHIFERR_NoDriver = 19
        CHIFERR_AccessDenied = 13
    else:
        CHIFERR_NoDriver = 2
        CHIFERR_AccessDenied = 5


class HpIloInitialError(Exception):
    """Raised when error during initialization of iLO Chif channel"""

    pass


class HpIloNoChifDriverError(Exception):
    """Raised when error during initialization of iLO Chif channel"""

    pass


class HpIloChifAccessDeniedError(Exception):
    """Raised when error during initialization of iLO Chif channel"""

    pass


class HpIloPrepareAndCreateChannelError(Exception):
    """Raised when error during initialization of iLO Chif channel"""

    pass


class HpIloChifPacketExchangeError(Exception):
    """Raised when errors encountered when exchanging chif packet"""

    pass


class HpIloReadError(Exception):
    """Raised when errors encountered when reading from iLO"""

    pass


class HpIloWriteError(Exception):
    """Raised when errors encountered when writing to iLO"""

    pass


class HpIloSendReceiveError(Exception):
    """Raised when errors encountered when reading form iLO after sending"""

    pass


class HpIloNoDriverError(Exception):
    """Raised when errors encountered when there is no ilo driver"""

    pass


class HpIlo(object):
    """Base class of interaction with iLO"""

    def __init__(self, dll=None, log_dir=None):
        fhandle = c_void_p()
        self.dll = dll
        self.log_dir = log_dir

        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug("Initializing with DLL: %s and log directory: %s", dll, log_dir)
            self.dll.enabledebugoutput.argtypes = [c_char_p]
            if log_dir is not None:
                logdir_c = create_string_buffer(log_dir.encode("UTF-8"))
                LOGGER.debug("Enabling debug output with log directory: %s", log_dir)
                self.dll.enabledebugoutput(logdir_c)

        try:
            LOGGER.debug("Calling ChifInitialize...")
            self.dll.ChifInitialize(None)

            self.dll.ChifCreate.argtypes = [c_void_p]
            self.dll.ChifCreate.restype = c_uint32

            LOGGER.debug("Calling ChifCreate...")
            status = self.dll.ChifCreate(byref(fhandle))
            if status != BlobReturnCodes.SUCCESS:
                LOGGER.error("Error occurred while creating channel. Error code: %s", status)
                raise HpIloInitialError(f"Error {status} occurred while trying to create a channel.")

            self.fhandle = fhandle
            LOGGER.debug("Channel created successfully with handle: %s", fhandle)

            if "skip_ping" not in os.environ:
                LOGGER.debug("Attempting to ping iLO...")
                status = self.dll.ChifPing(self.fhandle)
                if status != BlobReturnCodes.SUCCESS:
                    errmsg = f"Error {status} occurred while trying to open a channel to iLO"
                    if status == BlobReturnCodes.CHIFERR_NoDriver:
                        errmsg = "chif"
                    elif status == BlobReturnCodes.CHIFERR_AccessDenied:
                        errmsg = "You must be root/Administrator to use this program."
                    LOGGER.error("Ping failed: %s", errmsg)
                    raise HpIloInitialError(errmsg)

                LOGGER.debug("Ping to iLO successful.")

                # Setting receive timeout
                LOGGER.debug("Setting receive timeout to 60000ms...")
                self.dll.ChifSetRecvTimeout(self.fhandle, 60000)
                LOGGER.debug("Receive timeout set successfully.")
        except Exception as e:
            LOGGER.error("An error occurred during initialization: %s", str(e))
            raise

    def chif_packet_exchange(self, data):
        """Function for handling chif packet exchange

        :param data: data to be sent for packet exchange
        :type data: str

        """
        LOGGER.info("Starting chif packet exchange...")

        LOGGER.debug("Data ready to be sent...")
        # LOGGER.debug(f"Data to be sent: {data[:50]}... (first 50 bytes)")

        datarecv = self.dll.get_max_buffer_size()
        buff = create_string_buffer(bytes(data))
        recbuff = create_string_buffer(datarecv)

        try:
            error = self.dll.ChifPacketExchange(self.fhandle, byref(buff), byref(recbuff), datarecv)
        except Exception as e:
            LOGGER.error(f"Error occurred while calling ChifPacketExchange: {e}")
            raise

        if error != BlobReturnCodes.SUCCESS:
            if error == 8:
                raise HpIloChifPacketExchangeError(
                    "High security mode or Host Authentication has been enabled, Please provide valid credentials \n "
                )
            else:
                raise HpIloChifPacketExchangeError(f"Error {error} occurred while exchanging chif packet")

        LOGGER.info("Chif packet exchange successful.")

        # Process received packet
        pkt = bytearray()

        if datarecv is None:
            LOGGER.warning("Received buffer is empty or None.")
            pkt = bytearray(recbuff[:])
        else:
            pkt = bytearray(recbuff[:datarecv])
            LOGGER.debug("Received data...")
            # LOGGER.debug(f"Received data: {pkt[:50]}... (first 50 bytes)")

        return pkt

    def send_receive_raw(self, data, retries=10):
        """Function implementing proper send receive retry protocol

        :param data: data to be sent for packet exchange
        :type data: str
        :param retries: number of retries for reading data from iLO
        :type retries: int

        """
        tries = 0
        sequence = struct.unpack("<H", bytes(data[2:4]))[0]

        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug("Retries: %d, Initial sequence: %d", retries, sequence)

        while tries < retries:
            try:
                if LOGGER.isEnabledFor(logging.DEBUG):
                    LOGGER.debug("Attempt %d, sending data to iLO", tries + 1)

                resp = self.chif_packet_exchange(data)

                # Log the response data for debugging (consider limiting if large)
                if LOGGER.isEnabledFor(logging.DEBUG):
                    LOGGER.debug("Received response...")

                # Check for matching sequence
                received_sequence = struct.unpack("<H", bytes(resp[2:4]))[0]
                if sequence != received_sequence:
                    if LOGGER.isEnabledFor(logging.DEBUG):
                        LOGGER.debug(
                            "Attempt %d: Bad sequence number. Expected: %d, Got: %d",
                            tries + 1,
                            sequence,
                            received_sequence,
                        )
                    continue  # Retry with a new attempt

                # Successfully received response with correct sequence
                if LOGGER.isEnabledFor(logging.DEBUG):
                    LOGGER.debug("Attempt %d: Successful response with correct sequence.", tries + 1)

                return resp

            except Exception as excp:
                time.sleep(1)

                # Log exception on each retry attempt
                if LOGGER.isEnabledFor(logging.DEBUG):
                    LOGGER.debug("Attempt %d: Error while reading iLO: %s", tries + 1, str(excp))

                if tries == (retries - 1):
                    self.close()

                    if LOGGER.isEnabledFor(logging.DEBUG):
                        LOGGER.debug("Final attempt failed. Closing connection.")
                    raise excp  # Raise after final attempt

            tries += 1

        # If retries exhausted, log and raise custom exception
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug("Retries exhausted. iLO not responding after %d attempts.", retries)

        raise HpIloSendReceiveError("iLO not responding")

    def close(self):
        """Chif close function"""
        try:
            if self.fhandle is not None:
                LOGGER.debug("Calling ChifClose...")
                self.dll.ChifClose(self.fhandle)
                self.fhandle = None
        except Exception:
            pass

    def __del__(self):
        """Chif delete function"""
        self.close()



###
# Copyright 2024 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Base implementation for interaction with vnic interface"""

from redfish.rest.connections import ChifDriverMissingOrNotFound, VnicNotEnabledError
from redfish.rest.v1 import InvalidCredentialsError
import logging
from ctypes import (
    POINTER,
    byref,
    c_bool,
    c_char_p,
    c_int,
    c_uint8,
    c_wchar_p,
    create_string_buffer,
)

from redfish.hpilo.risblobstore2 import BlobStore2
from redfish.hpilo.rishpilo import BlobReturnCodes


# ---------End of imports---------
# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


class GenerateAndSaveAccountError(Exception):
    """Raised when errors occured while generating and saving app account"""

    pass


class ReactivateAppAccountTokenError(Exception):
    """Raised when errors occurred while reactivating app account"""

    pass

class InactiveAppAccountTokenError(Exception):
    """Raised when inactive app account token"""

    pass


class RemoveAccountError(Exception):
    """Raised when errors occured while removing app account"""

    pass


class AppAccountExistsError(Exception):
    """Raised when errors occured while removing app account"""

    pass


class VnicLoginError(Exception):
    """Raised when error occurs while VNIC login"""

    pass


class AppIdListError(Exception):
    """Raised when error occurs during CompareAppIds"""

    pass


class SavinginTPMError(Exception):
    """Raised when error occurs while saving app account in TPM"""

    pass


class SavinginiLOError(Exception):
    """Raised when error occurs while saving app account in iLO"""

    pass


class GenBeforeLoginError(Exception):
    """Raised when error occurs while getting iLO Gen before login"""

    pass


class InvalidCommandLineError(Exception):
    """Raised when invalid command line arguments are passed"""

    pass


class AppAccount(object):
    def __init__(self, appname=None, appid=None, salt=None, username=None, password=None, log_dir=None):
        """Initialize the AppAccount object and configure logging."""
        self.appname = (appname or "self_register").encode("utf-8")
        self.dll = BlobStore2.gethprestchifhandle()
        self.salt = (salt or "self_register").encode("utf-8")
        self.username = username
        self.password = password
        self.log_dir = log_dir

        LOGGER.debug(
            f"Initialization parameters - AppName: {self.appname.decode('utf-8')}, "
            f"Salt: {self.salt.decode('utf-8')}, Username: {self.username if self.username else 'None'}"
        )

        if log_dir and LOGGER.isEnabledFor(logging.DEBUG):
            logdir_c = create_string_buffer(log_dir.encode("utf-8"))
            self.dll.enabledebugoutput(logdir_c)
            LOGGER.info("Debug output enabled. Logs stored in: %s", log_dir)

        if appid and len(appid) == 4:
            appid = self.ExpandAppId(appid)

        self.appid = (appid or "self_register").encode("utf-8")

    def ExpandAppId(self, appid):
        """Calls the DLL function to expand a 4-character app ID."""
        self.dll.ExpandAppid.argtypes = [c_char_p, POINTER(c_char_p)]
        self.dll.ExpandAppid.restype = c_int

        host_app_id_c = c_char_p(appid.encode("utf-8"))
        expanded_app_id_c = c_char_p(None)

        LOGGER.debug(f"Calling ExpandAppid with AppID: {appid}")
        ret = self.dll.ExpandAppid(host_app_id_c, byref(expanded_app_id_c))
        if ret:
            LOGGER.error(f"Failed to expand App ID {appid} (Return code: {ret})")
            raise AppAccountExistsError()

        expanded_id = expanded_app_id_c.value.decode("utf-8") if expanded_app_id_c else None
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(f"Successfully expanded App ID {appid} to {expanded_id}")
        return expanded_id

    def generate_and_save_apptoken(self):
        """Generate and save an application account, logging each step."""
        LOGGER.info("Starting generate_and_save_apptoken process.")

        self.dll.GenerateAppToken.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p]
        self.dll.GenerateAppToken.restype = c_int

        self.username = self.username.encode("utf-8")
        self.password = self.password.encode("utf-8")

        LOGGER.debug("Calling GenerateAppToken for AppID: %s", self.appid.decode("utf-8"))

        returncode = self.dll.GenerateAppToken(self.appid, self.appname, self.salt, self.username, self.password)

        if returncode == BlobReturnCodes.SUCCESS:
            LOGGER.info("Application account successfully generated and saved.")
            return BlobReturnCodes.SUCCESS
        elif returncode == -54:
            LOGGER.error("Failed to save account in TPM.")
            raise SavinginTPMError()
        elif returncode == -53:
            LOGGER.error("Failed to save account in iLO.")
            raise SavinginiLOError()
        elif returncode == -55:
            LOGGER.warning("Application account already exists.")
            raise AppAccountExistsError()
        elif returncode == -8:
            raise InvalidCredentialsError(0)
        else:
            LOGGER.critical("Unknown return code: %d", returncode)
            raise GenerateAndSaveAccountError()

    def reactivate_apptoken(self):
        """Reactivate an application account, logging each step."""
        LOGGER.info("Starting reactivate_apptoken process.")

        self.dll.ReactivateAppToken.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p]
        self.dll.ReactivateAppToken.restype = c_int

        user_name = self.username.encode("utf-8")
        password = self.password.encode("utf-8")

        LOGGER.debug("Calling ReactivateAppToken for AppID: %s", self.appid.decode("utf-8"))

        return_code = self.dll.ReactivateAppToken(self.appid, self.appname, self.salt, user_name, password)

        if return_code == BlobReturnCodes.SUCCESS:
            LOGGER.info("Application account has been reactivated successfully.")
            return BlobReturnCodes.SUCCESS
        elif return_code == -8:
            LOGGER.error("Please enter valid credentials.")
            raise InvalidCredentialsError(0)
        else:
            LOGGER.critical("Unknown return code: %d", return_code)
            raise ReactivateAppAccountTokenError()

    def remove_apptoken(self):
        returncode = 1
        if self.username and self.password:
            LOGGER.debug(f"Using credentials to remove for AppID: {self.appid.decode('utf-8')}")
            self.dll.DeleteAppTokenUsingCreds.argtypes = [c_char_p, c_char_p, c_char_p]
            self.dll.DeleteAppTokenUsingCreds.restype = c_int
            self.username = self.username.encode("utf-8")
            self.password = self.password.encode("utf-8")
            returncode = self.dll.DeleteAppTokenUsingCreds(self.appid, self.username, self.password)

        elif self.appname and self.salt:
            LOGGER.debug(f"Using app details to remove for AppID : {self.appid.decode('utf-8')}")
            self.dll.DeleteAppToken.argtypes = [c_char_p, c_char_p, c_char_p]
            self.dll.DeleteAppToken.restype = c_int
            returncode = self.dll.DeleteAppToken(self.appid, self.appname, self.salt)

        LOGGER.debug(f"Delete operation returned code: {returncode}")

        if returncode == BlobReturnCodes.SUCCESS:
            return BlobReturnCodes.SUCCESS
        elif returncode == -8:
            raise InvalidCredentialsError(0)
        else:
            raise RemoveAccountError("Error occurred while removing app account.")

    def apptoken_exists(self):
        """Check if an application account exists in TPM."""
        LOGGER.info("Checking if application account exists.")

        self.dll.AppIdExistsinTPM.argtypes = [c_char_p, POINTER(c_bool)]
        self.dll.AppIdExistsinTPM.restype = c_int

        vexists = c_bool(False)
        LOGGER.debug(f"Checking existence for AppID: {self.appid.decode('utf-8')}")
        returncode = self.dll.AppIdExistsinTPM(self.appid, byref(vexists))
        LOGGER.debug(f"AppIdExistsinTPM returned: {returncode}")

        if returncode == BlobReturnCodes.SUCCESS:
            LOGGER.info("app account existence check completed. Exists: %s", vexists.value)
            return vexists.value
        else:
            LOGGER.error("Failed to check account existence")
            LOGGER.debug(f"Failed with return code: {returncode}")
            raise AppAccountExistsError()

    def vlogin(self):
        LOGGER.info("Starting VNIC login process.")

        self.dll.GetSessionID.argtypes = [c_char_p, c_char_p, c_char_p, POINTER(c_char_p), POINTER(c_char_p)]
        self.dll.GetSessionID.restype = c_int
        session_id = c_char_p()
        session_location = c_char_p()
        errorcode = 0

        LOGGER.debug(f"Attempting login for AppID: {self.appid.decode('utf-8')}")
        try:
            errorcode = self.dll.GetSessionID(
                self.appid, self.appname, self.salt, byref(session_id), byref(session_location)
            )
            LOGGER.debug(f"GetSessionID returned code: {errorcode}")
            if errorcode == 2:
                raise InactiveAppAccountTokenError()
            elif errorcode != BlobReturnCodes.SUCCESS:
                raise Exception
            session_id = session_id.value.decode()
            session_location = session_location.value.decode()
            session_location = "/" + session_location.split("/", 3)[-1]
            LOGGER.debug("Successfully decoded session information")
            LOGGER.info(
                "VNIC login successful. Session Location: %s, Session ID: ****%s", session_location, session_id[-4:]
            )

            return session_id, session_location
        except InactiveAppAccountTokenError:
            LOGGER.error("Login failed due to an inactive or expired App Account token.")
            raise InactiveAppAccountTokenError()
        except Exception as excp:
            LOGGER.error("VNIC login failed: %s", str(excp))
            raise VnicLoginError()

    def vnic_exists(self):
        """Check if the VNIC is ready for use."""
        LOGGER.info("Checking if VNIC is ready for use.")

        self.dll.ReadyToUse.restype = c_int
        LOGGER.debug("Calling ReadyToUse")
        status = self.dll.ReadyToUse()
        LOGGER.debug(f"ReadyToUse returned status: {status}")

        if status == BlobReturnCodes.SUCCESS:
            LOGGER.info("VNIC is ready for use.")
            return True
        else:
            LOGGER.warning("VNIC is not ready for use.")
            return False

    def version_beforelogin(self):
        """Detect the iLO version before login."""
        LOGGER.info("Detecting iLO version before login.")

        self.dll.DetectILO.argtypes = [c_char_p, POINTER(c_int), POINTER(c_int)]
        self.dll.DetectILO.restype = c_int

        ilo_type = c_int(0)
        security_state = c_int(0)

        if not self.appid:
            self.appid = "self_register".encode("utf-8")

        returncode = self.dll.DetectILO(self.appid, byref(ilo_type), byref(security_state))
        LOGGER.debug(f"DetectILO returned: type={ilo_type.value}, security={security_state.value}, code={returncode}")

        if returncode == BlobReturnCodes.SUCCESS:
            LOGGER.info(
                "iLO version successfully detected: %d, Security State: %d", ilo_type.value, security_state.value
            )
            return ilo_type.value, security_state.value
        elif returncode == -1 and ilo_type.value == 100:
            LOGGER.info(
                "Detectilo returned ilotype: %d, Security State: %d, " "returncode: %d",
                ilo_type.value,
                security_state.value,
                returncode,
            )
            raise ChifDriverMissingOrNotFound()
        elif returncode == -50 and ilo_type.value == 101:
            LOGGER.info(
                "Detectilo returned ilotype: %d, Security State: %d, " "returncode: %d",
                ilo_type.value,
                security_state.value,
                returncode,
            )
            raise VnicNotEnabledError()
        elif returncode == -2:
            LOGGER.info("Invalid parameters were passed in the command line. Detectilo iLO returncode: -2")
            raise InvalidCommandLineError()
        else:
            LOGGER.info(
                "Detectilo returned ilotype: %d, Security State: %d, " "returncode: %d",
                ilo_type.value,
                security_state.value,
                returncode,
            )
            LOGGER.error("Failed to detect iLO version.")
            raise GenBeforeLoginError()

    def GetIPAddress(self):
        """Retrieve the IP address of the iLO interface."""
        LOGGER.info("Retrieving IP address.")

        LOGGER.debug("Calling GetIPAddress")
        self.dll.GetIPAddress.restype = c_wchar_p
        ip_addr = self.dll.GetIPAddress()
        LOGGER.debug(f"GetIPAddress returned: {ip_addr if ip_addr else 'None'}")

        LOGGER.info("IP address retrieved: %s", ip_addr)
        return ip_addr

    def CompareAppIds(self):
        """Compare application IDs and return a structured dictionary."""
        LOGGER.info("Starting CompareAppIds process.")

        self.dll.CompareAppIds.argtypes = [
            c_char_p,
            c_char_p,
            c_char_p,
            POINTER(POINTER(c_char_p)),
            POINTER(POINTER(c_char_p)),
            POINTER(c_int),
            POINTER(POINTER(c_uint8)),
            POINTER(POINTER(c_uint8)),
        ]
        self.dll.CompareAppIds.restype = c_int

        app_ids_ptr = POINTER(c_char_p)()
        app_names_ptr = POINTER(c_char_p)()
        app_id_count = c_int()
        in_tpm_ptr = POINTER(c_uint8)()
        in_ilo_ptr = POINTER(c_uint8)()

        LOGGER.debug(
            "Calling CompareAppIds with AppID: %s, AppName: %s",
            self.appid.decode("utf-8"),
            self.appname.decode("utf-8"),
        )

        returncode = self.dll.CompareAppIds(
            self.appid,
            self.appname,
            self.salt,
            byref(app_ids_ptr),
            byref(app_names_ptr),
            byref(app_id_count),
            byref(in_tpm_ptr),
            byref(in_ilo_ptr),
        )
        if returncode != BlobReturnCodes.SUCCESS:
            LOGGER.error("Error occurred in CompareAppIds. Return Code: %d", returncode)
            raise AppIdListError("Error occurred while retrieving App Id information.\n")

        LOGGER.info("CompareAppIds executed successfully. Number of App IDs: %d", app_id_count.value)
        app_ids = [app_ids_ptr[i].decode("utf-8") for i in range(app_id_count.value)]
        app_names = [app_names_ptr[i].decode("utf-8") for i in range(app_id_count.value)]
        in_tpm_list = [bool(in_tpm_ptr[i]) for i in range(app_id_count.value)]
        in_ilo_list = [bool(in_ilo_ptr[i]) for i in range(app_id_count.value)]

        appid_dict = list()
        for i in range(app_id_count.value):
            appid_dict.append(
                {
                    "ApplicationID": app_ids[i],
                    "ApplicationName": app_names[i],
                    "ExistsInTPM": in_tpm_list[i],
                    "ExistsIniLO": in_ilo_list[i],
                }
            )

        LOGGER.info("CompareAppIds completed successfully. Processed %d AppIDs.", len(appid_dict))

        return appid_dict



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Base implementation for interaction with blob store interface"""

# ---------Imports---------

import os
import sys
import platform
import sysconfig
import logging
import random
import struct
import string
import time
import ctypes
from ctypes import POINTER, c_char_p, c_ubyte, c_uint, c_ushort, c_void_p, create_string_buffer

from redfish.hpilo.rishpilo import BlobReturnCodes as hpiloreturncodes
from redfish.hpilo.rishpilo import (
    HpIlo,
    HpIloChifPacketExchangeError,
    HpIloInitialError,
)

if os.name == "nt":
    from ctypes import windll
else:
    from _ctypes import dlclose

# Flags from <dlfcn.h> (Linux/Unix)
RTLD_LAZY = 0x00001
RTLD_GLOBAL = 0x00100

# ---------End of imports---------
# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)

# ---------End of debug logger---------
# -----------------------Error Returns----------------------


class UnexpectedResponseError(Exception):
    """Raise when we get data that we don't expect from iLO"""

    pass


class HpIloError(Exception):
    """Raised when iLO returns non-zero error code"""

    pass


class Blob2CreateError(Exception):
    """Raised when create operation fails"""

    pass


class Blob2InfoError(Exception):
    """Raised when create operation fails"""

    pass


class Blob2ReadError(Exception):
    """Raised when read operation fails"""

    pass


class Blob2WriteError(Exception):
    """Raised when write operation fails"""

    pass


class Blob2DeleteError(Exception):
    """Raised when delete operation fails"""

    pass


class Blob2OverrideError(Exception):
    """Raised when delete operation fails because of it been overwritten"""

    pass


class BlobRetriesExhaustedError(Exception):
    """Raised when max retries have been attempted for same operation"""

    pass


class Blob2FinalizeError(Exception):
    """Raised when finalize operation fails"""

    pass


class Blob2ListError(Exception):
    """Raised when list operation fails"""

    pass


class Blob2SecurityError(Exception):
    """Raised when there is an issue with security"""

    pass


class BlobNotFoundError(Exception):
    """Raised when blob not found in key/namespace"""

    pass


class ChifDllMissingError(Exception):
    """Raised when unable to obtain ilorest_chif dll handle"""

    pass


class EncryptionEnabledError(Exception):
    """Raised when high security encryption is enabled"""

    pass


# ----------------------------------------------------------

# -------------------Helper functions-------------------------


class BlobReturnCodes(object):
    """Blob store return codes.

    SUCCESS           success
    BADPARAMETER      bad parameter supplied
    NOTFOUND          blob name not found
    NOTMODIFIED       call did not perform the operation

    """

    SUCCESS = 0
    BADPARAMETER = 2
    NOTFOUND = 12
    NOTMODIFIED = 20


class BlobStore2(object):
    """Blob store 2 class"""

    def __init__(self, log_dir=None):
        lib = self.gethprestchifhandle()
        self.log_dir = log_dir
        self.channel = HpIlo(dll=lib, log_dir=log_dir)
        self.max_retries = 44
        self.max_read_retries = 3
        self.delay = 0.25

        LOGGER.info("BlobStore initialized with log directory: %s", log_dir)

    def __del__(self):
        """Blob store 2 close channel function"""
        if hasattr(self, "channel"):
            self.channel.close()

    def create(self, key, namespace):
        """
        Create the blob.

        :param key: The blob key to create.
        :type key: str
        :param namespace: The blob namespace to create the key in.
        :type namespace: str
        :return: The response from the blob creation request.
        :rtype: bytearray
        :raises HpIloError: If the operation fails.
        """
        LOGGER.info("Attempting to create blob: Key=%s, Namespace=%s", key, namespace)

        try:
            lib = self.gethprestchifhandle()
            lib.create_not_blobentry.argtypes = [c_char_p, c_char_p]
            lib.create_not_blobentry.restype = POINTER(c_ubyte)

            name = create_string_buffer(key.encode("utf-8"))
            namespace = create_string_buffer(namespace.encode("utf-8"))

            LOGGER.debug("Calling create_not_blobentry()")
            ptr = lib.create_not_blobentry(name, namespace)

            LOGGER.debug("Preparing request data")
            data = ptr[: lib.size_of_createRequest()]
            data = bytearray(data)

            LOGGER.info("Sending blob creation request")
            resp = self._send_receive_raw(data)

            errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
            if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
                LOGGER.error("Blob creation failed. Error Code: %d", errorcode)
                raise HpIloError(errorcode)

            LOGGER.info("Blob creation successful for Key=%s, Namespace=%s", key, namespace)
            return resp

        except Exception as e:
            LOGGER.exception("Exception during blob creation: %s", str(e))
            raise

        finally:
            LOGGER.debug("Unloading library handle")
            self.unloadchifhandle(lib)

    def get_info(self, key, namespace, retries=0):
        """Get information for a particular blob.

        :param key: The blob key to retrieve.
        :type key: str.
        :param namespace: The blob namespace to retrieve the key from.
        :type namespace: str.

        """
        LOGGER.info(f"get_info called with key='{key}', namespace='{namespace}', retries={retries}")

        lib = self.gethprestchifhandle()
        lib.get_info.argtypes = [c_char_p, c_char_p]
        lib.get_info.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namspace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.get_info(name, namspace)
        data = ptr[: lib.size_of_infoRequest()]
        data = bytearray(data)
        delay = self.delay
        while retries <= self.max_retries:
            LOGGER.debug(f"Attempt {retries+1}/{self.max_retries} - Sending request to iLO.")

            resp = self._send_receive_raw(data)
            errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
            header = resp[:8].hex()

            LOGGER.debug(f"Response received - Error Code: {errorcode}, Headers: {header}, Size: {len(resp)} bytes")

            if errorcode == BlobReturnCodes.BADPARAMETER:
                if retries < self.max_retries:
                    LOGGER.warning(
                        f"BADPARAMETER error received. Retrying in {delay} seconds... "
                        f"({retries+1}/{self.max_retries})"
                    )
                    time.sleep(delay)
                    retries += 1
                    delay += 0.05
                    continue
                else:
                    LOGGER.error(f"Max retries ({self.max_retries}) exceeded. Raising Blob2OverrideError.")
                    raise Blob2OverrideError(errorcode)

            elif errorcode == BlobReturnCodes.NOTFOUND:
                LOGGER.error(f"BlobNotFoundError: key='{key}', namespace='{namespace}'")
                raise BlobNotFoundError(key, namespace)

            elif errorcode not in (BlobReturnCodes.SUCCESS, BlobReturnCodes.NOTMODIFIED):
                LOGGER.error(f"HpIloError: Unexpected error code {errorcode}.")
                raise HpIloError(errorcode)

            else:
                LOGGER.info("Request successful. Extracting response data.")
                break

        response = resp[lib.size_of_responseHeaderBlob() :]
        self.unloadchifhandle(lib)

        LOGGER.info(f"get_info completed successfully for key='{key}', namespace='{namespace}'.")
        return response

    def read(self, key, namespace, retries=0):
        """Read a particular blob

        :param key: The blob key to be read.
        :type key: str.
        :param namespace: The blob namespace to read the key from.
        :type namespace: str.

        """
        LOGGER.info("Starting blob read: Key=%s, Namespace=%s, Retry=%d", key, namespace, retries)

        try:
            lib = self.gethprestchifhandle()
            maxread = lib.max_read_size()
            readsize = lib.size_of_readRequest()
            readhead = lib.size_of_responseHeaderBlob()
            LOGGER.debug(f"Read parameters: maxread={maxread}, readsize={readsize}, readhead={readhead}")
            self.unloadchifhandle(lib)

            # Get blob metadata
            LOGGER.debug("Fetching blob info to determine size.")
            blob_info = self.get_info(key, namespace)
            blobsize = struct.unpack("<I", bytes(blob_info[0:4]))[0]
            LOGGER.info("Blob size determined: %d bytes", blobsize)

            bytes_read = 0
            data = bytearray()

            while bytes_read < blobsize:
                count = min(maxread - readsize, blobsize - bytes_read)

                LOGGER.debug("Reading fragment: Offset=%d, Size=%d", bytes_read, count)
                recvpkt = self.read_fragment(key, namespace, bytes_read, count)

                newreadsize = readhead + 4
                bytesread = struct.unpack("<I", bytes(recvpkt[readhead:newreadsize]))[0]
                LOGGER.debug("Bytes read in current fragment: %d", bytesread)

                if bytesread == 0:
                    LOGGER.debug(f"Zero bytes read in fragment. Fragment metadata: offset={bytes_read}, count={count}")
                    if retries < self.max_read_retries:
                        LOGGER.warning(
                            "Read attempt failed. Retrying (Attempt %d/%d).", retries + 1, self.max_read_retries
                        )
                        return self.read(key=key, namespace=namespace, retries=retries + 1)
                    else:
                        LOGGER.error("Maximum read retries (%d) exceeded.", self.max_read_retries)
                        raise BlobRetriesExhaustedError()

                data.extend(recvpkt[newreadsize : newreadsize + bytesread])
                bytes_read += bytesread
                LOGGER.debug("Total bytes read so far: %d/%d", bytes_read, blobsize)

            LOGGER.info("Blob read successfully: Key=%s, Namespace=%s, Total Bytes=%d", key, namespace, len(data))
            return data

        except Exception as e:
            LOGGER.exception("Error during blob read: %s", str(e))
            raise

    def read_fragment(self, key, namespace, offset=0, count=1):
        """Fragmented version of read function for large blobs

        :param key: The blob key to be read.
        :type key: str.
        :param namespace: The blob namespace to read the key from.
        :type namespace: str.
        :param offset: The data offset for the current fragmented read.
        :type key: int.
        :param count: The data count for the current fragmented read.
        :type namespace: int.

        """
        LOGGER.info(
            "Starting fragmented blob read: Key=%s, Namespace=%s, Offset=%d, Count=%d", key, namespace, offset, count
        )

        try:
            lib = self.gethprestchifhandle()
            lib.read_fragment.argtypes = [c_uint, c_uint, c_char_p, c_char_p]
            lib.read_fragment.restype = POINTER(c_ubyte)

            name = create_string_buffer(key.encode("utf-8"))
            namespace_buf = create_string_buffer(namespace.encode("utf-8"))

            LOGGER.debug("Preparing request data for read_fragment.")
            ptr = lib.read_fragment(offset, count, name, namespace_buf)

            request_size = lib.size_of_readRequest()
            response_size = lib.size_of_readResponse()

            LOGGER.debug("Extracting request data (Size=%d)", request_size)
            data = bytearray(ptr[:request_size])

            LOGGER.info("Sending fragmented read request to iLO.")
            resp = self._send_receive_raw(data)

            # Check response error code
            errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
            LOGGER.debug(
                f"Response received - Headers: Status={errorcode}, Size={len(resp)}, Expected Size={response_size}"
            )
            if len(resp) < response_size:
                LOGGER.warning("Received response smaller than expected. Padding response.")
                resp = resp + b"\0" * (response_size - len(resp))

            LOGGER.info("Fragmented blob read successful: Key=%s, Bytes Received=%d", key, len(resp))
            return resp

        except Exception as e:
            LOGGER.exception("Error occurred during read_fragment: %s", str(e))
            raise

    def write(self, key, namespace, data=None):
        """Write a particular blob

        :param key: The blob key to be written.
        :type key: str.
        :param namespace: The blob namespace to write the key in.
        :type namespace: str.
        :param data: The blob data to be written.
        :type data: str.

        """
        LOGGER.info(f"Starting to write blob with key: {key} in namespace: {namespace}")

        lib = self.gethprestchifhandle()
        maxwrite = lib.max_write_size()
        writesize = lib.size_of_writeRequest()

        LOGGER.debug(f"Max write size: {maxwrite}, Write request size: {writesize}")

        self.unloadchifhandle(lib)

        if data:
            data_length = len(data)
            bytes_written = 0
            LOGGER.info(f"Data length: {data_length}")

            while bytes_written < data_length:
                if (maxwrite - writesize) < (data_length - bytes_written):
                    count = maxwrite - writesize
                else:
                    count = data_length - bytes_written

                write_blob_size = bytes_written
                LOGGER.debug(f"Writing data from byte {write_blob_size} to {write_blob_size + count}")

                try:
                    self.write_fragment(
                        key,
                        namespace=namespace,
                        data=data[write_blob_size : write_blob_size + count],
                        offset=write_blob_size,
                        count=count,
                    )
                except Exception as e:
                    LOGGER.error(f"Failed to write fragment for blob with key {key} in namespace {namespace}: {e}")
                    raise

                bytes_written += count
                LOGGER.debug(f"{bytes_written} bytes written so far.")

        try:
            result = self.finalize(key, namespace=namespace)
            LOGGER.info(f"Finalization successful for key: {key} in namespace: {namespace}")
            return result
        except Exception as e:
            LOGGER.error(f"Failed to finalize the write for blob with key {key} in namespace {namespace}: {e}")
            raise

    def write_fragment(self, key, namespace, data=None, offset=0, count=1):
        """Fragmented version of write function for large blobs

        :param key: The blob key to be written.
        :type key: str.
        :param namespace: The blob namespace to write the key in.
        :type namespace: str.
        :param data: The blob data to be written to blob.
        :type data: str.
        :param offset: The data offset for the current fragmented write.
        :type key: int.
        :param count: The data count for the current fragmented write.
        :type count: int.

        """
        LOGGER.info(f"Starting fragmented write for blob with key: {key} in namespace: {namespace}")
        LOGGER.debug(f"Fragment write params: offset={offset}, count={count}")

        lib = self.gethprestchifhandle()
        lib.write_fragment.argtypes = [c_uint, c_uint, c_char_p, c_char_p]
        lib.write_fragment.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namespace = create_string_buffer(namespace.encode("utf-8"))

        # Call the external library function
        try:
            LOGGER.debug(f"Write fragment request: key={key}, namespace={namespace}, offset={offset}, count={count}")
            ptr = lib.write_fragment(offset, count, name, namespace)
        except Exception as e:
            LOGGER.error(f"Error while calling write_fragment for key: {key} in namespace: {namespace}: {e}")
            raise

        LOGGER.debug(f"Preparing write fragment data: request_size={count} bytes")

        sendpacket = ptr[: lib.size_of_writeRequest()]

        if isinstance(data, str):
            data = data.encode("utf-8")

        dataarr = bytearray(sendpacket)
        dataarr.extend(memoryview(data))
        LOGGER.debug(f"Request metadata: offset={offset}, size={count}")

        # Send the data
        try:
            resp = self._send_receive_raw(dataarr)
        except Exception as e:
            LOGGER.error(f"Error during sending/receiving raw data for blob with key: {key}: {e}")
            raise

        # Check for errors in response
        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            LOGGER.error(f"Write fragment failed for key: {key} in namespace: {namespace}. Error code: {errorcode}")
            raise HpIloError(errorcode)

        LOGGER.info(f"Fragmented write successful for key: {key} in namespace: {namespace}")

        self.unloadchifhandle(lib)

        return resp

    def delete(self, key, namespace, retries=0):
        """Delete the blob.

        :param key: The blob key to be deleted.
        :type key: str
        :param namespace: The blob namespace to delete the key from.
        :type namespace: str
        :param retries: The number of retries if deletion fails (default 0).
        :type retries: int
        """
        LOGGER.info(f"Starting blob deletion: key={key}, namespace={namespace}, retries={retries}")

        lib = self.gethprestchifhandle()
        lib.delete_blob.argtypes = [c_char_p, c_char_p]
        lib.delete_blob.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode("utf-8"))
        namspace = create_string_buffer(namespace.encode("utf-8"))

        ptr = lib.delete_blob(name, namspace)
        data = ptr[: lib.size_of_deleteRequest()]
        data = bytearray(data)
        delay = self.delay
        while retries <= self.max_retries:
            LOGGER.debug(f"Attempt {retries + 1} of {self.max_retries}: Sending raw delete request.")

            resp = self._send_receive_raw(data)

            # Checking for errors in the received response
            errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
            LOGGER.debug(f"Received error code: {errorcode}")

            if errorcode == BlobReturnCodes.BADPARAMETER:
                LOGGER.warning(
                    f"BADPARAMETER error received (retries left: {self.max_retries - retries}). "
                    f"Retrying after {delay} seconds."
                )
                if retries < self.max_retries:
                    time.sleep(delay)
                    retries += 1
                    delay += 0.05
                    continue
                else:
                    LOGGER.warning(f"Max retries reached. Unable to delete blob key={key}. Raising Blob2OverrideError.")
                    raise Blob2OverrideError(errorcode)
            elif errorcode not in (BlobReturnCodes.SUCCESS, BlobReturnCodes.NOTMODIFIED):
                LOGGER.debug(f"Error during blob deletion: {errorcode}. Raising HpIloError.")
                raise HpIloError(errorcode)
            else:
                LOGGER.info(f"Blob deletion successful for key={key} in namespace={namespace}.")
                break

        self.unloadchifhandle(lib)

        LOGGER.info(f"Exiting delete function for key={key}. Final error code: {errorcode}")
        return errorcode

    def list(self, namespace):
        """List operation to retrieve all blobs in a given namespace.

        :param namespace: The blob namespace to retrieve the keys from.
        :type namespace: str.
        """
        LOGGER.info(f"Starting blob listing operation for namespace: {namespace}")

        lib = self.gethprestchifhandle()
        lib.list_blob.argtypes = [c_char_p]
        lib.list_blob.restype = POINTER(c_ubyte)

        namespace = create_string_buffer(namespace.encode("utf-8"))

        LOGGER.debug("Sending list blob request to iLO.")
        ptr = lib.list_blob(namespace)
        data = ptr[: lib.size_of_listRequest()]
        data = bytearray(data)

        # Send and receive raw data
        resp = self._send_receive_raw(data)

        # Check for errors in the response
        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        LOGGER.debug(f"Received error code: {errorcode}")

        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            LOGGER.error(f"Error during blob listing: {errorcode}. Raising HpIloError.")
            raise HpIloError(errorcode)

        # Padding response to match expected size
        resp = resp + b"\0" * (lib.size_of_listResponse() - len(resp))

        # Unload the handler
        self.unloadchifhandle(lib)

        LOGGER.info(f"Blob listing successful for namespace: {namespace}")
        return resp

    def finalize(self, key, namespace):
        """Finalize the blob.

        :param key: The blob key to be finalized.
        :type key: str.
        :param namespace: The blob namespace to finalize the key in.
        :type namespace: str.
        """
        LOGGER.info(f"Starting blob finalize operation for key: {key}, namespace: {namespace}")

        # Getting the handler library
        lib = self.gethprestchifhandle()
        lib.finalize_blob.argtypes = [c_char_p, c_char_p]
        lib.finalize_blob.restype = POINTER(c_ubyte)

        # Preparing request data
        name = create_string_buffer(key.encode("utf-8"))
        namespace = create_string_buffer(namespace.encode("utf-8"))

        # Sending the finalize request
        LOGGER.debug("Sending finalize blob request to iLO.")
        ptr = lib.finalize_blob(name, namespace)
        data = ptr[: lib.size_of_finalizeRequest()]
        data = bytearray(data)

        # Send and receive raw data
        resp = self._send_receive_raw(data)

        # Checking for errors in the response
        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        LOGGER.debug(f"Received error code: {errorcode}")

        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            LOGGER.error(f"Error during blob finalize: {errorcode}. Raising HpIloError.")
            raise HpIloError(errorcode)

        # Unload the handler
        self.unloadchifhandle(lib)

        LOGGER.info(f"Blob finalize successful for key: {key}, namespace: {namespace}")
        return errorcode

    def rest_immediate(
        self,
        req_data,
        rqt_key="RisRequest",
        rsp_key="RisResponse",
        rsp_namespace="volatile",
    ):
        """Read/write blob via immediate operation

        :param req_data: The blob data to be read/written.
        :type req_data: str.
        :param rqt_key: The blob key to be used for the request data.
        :type rqt_key: str.
        :param rsp_key: The blob key to be used for the response data.
        :type rsp_key: str.
        :param rsp_namespace: The blob namespace to retrieve the response from.
        :type rsp_namespace: str.
        """
        # Log generated keys
        rqt_key = "".join(random.choice(string.ascii_letters + string.digits) for _ in range(10))
        rsp_key = "".join(random.choice(string.ascii_letters + string.digits) for _ in range(10))
        LOGGER.debug(f"Generated request key: {rqt_key}, response key: {rsp_key}")
        LOGGER.debug(f"Request data size: {len(req_data)} bytes")

        lib = self.gethprestchifhandle()

        # Log request size check
        if len(req_data) < (lib.size_of_restImmediateRequest() + lib.max_write_size()):
            LOGGER.debug("Proceeding with immediate request (no blob required).")

            # Rest immediate operation
            lib.rest_immediate.argtypes = [c_uint, c_char_p, c_char_p]
            lib.rest_immediate.restype = POINTER(c_ubyte)

            name = create_string_buffer(rsp_key.encode("utf-8"))
            namespace = create_string_buffer(rsp_namespace.encode("utf-8"))

            ptr = lib.rest_immediate(len(req_data), name, namespace)
            sendpacket = ptr[: lib.size_of_restImmediateRequest()]
            mode = False
        else:
            # Log that we're handling large data
            LOGGER.debug("Data is large, proceeding with create/write operations.")

            self.create(rqt_key, rsp_namespace)
            self.write(rqt_key, rsp_namespace, req_data)

            lib.rest_immediate_blobdesc.argtypes = [c_char_p, c_char_p, c_char_p]
            lib.rest_immediate_blobdesc.restype = POINTER(c_ubyte)

            name = create_string_buffer(rqt_key.encode("utf-8"))
            namespace = create_string_buffer(rsp_namespace.encode("utf-8"))
            rspname = create_string_buffer(rsp_key.encode("utf-8"))

            ptr = lib.rest_immediate_blobdesc(name, rspname, namespace)
            sendpacket = ptr[: lib.size_of_restBlobRequest()]
            mode = True

        data = bytearray(sendpacket)

        if not mode:
            data.extend(req_data)

        LOGGER.debug("Sending request data to iLO.")

        # Sending the data
        resp = self._send_receive_raw(data)

        # Log response processing
        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        LOGGER.debug("Received error code: %s", errorcode)

        if errorcode == BlobReturnCodes.NOTFOUND:
            LOGGER.error("Blob not found for key: %s in namespace: %s", rsp_key, rsp_namespace)
            raise BlobNotFoundError(rsp_key, rsp_namespace)

        recvmode = struct.unpack("<I", bytes(resp[12:16]))[0]

        fixdlen = lib.size_of_restResponseFixed()
        response = resp[fixdlen : struct.unpack("<I", bytes(resp[16:20]))[0] + fixdlen]

        tmpresponse = None
        if errorcode == BlobReturnCodes.SUCCESS and not mode:
            if recvmode == 0:
                tmpresponse = "".join(map(chr, response))
                LOGGER.debug("Successfully received response: %s", tmpresponse)
        elif errorcode == BlobReturnCodes.NOTMODIFIED and not mode:
            if recvmode == 0:
                tmpresponse = "".join(map(chr, response))
                LOGGER.debug("No modification detected, response: %s", tmpresponse)
        elif errorcode == BlobReturnCodes.SUCCESS:
            if recvmode == 0:
                tmpresponse = "".join(map(chr, response))
                LOGGER.debug("Successfully received response after modification: %s", tmpresponse)
        elif recvmode == 0:
            LOGGER.error("Error processing response: %s", errorcode)
            raise HpIloError(errorcode)

        # Cleanup and response handling
        self.unloadchifhandle(lib)

        if not tmpresponse and recvmode == 1:
            tmpresponse = self.read(rsp_key, rsp_namespace)
            LOGGER.debug("Fallback response: %s", tmpresponse)

            try:
                self.delete(rsp_key, rsp_namespace)
                LOGGER.debug("Successfully deleted blob with key: %s", rsp_key)
            except Exception as excp:
                LOGGER.warning("Error deleting blob: %s", excp)
                raise excp
        else:
            try:
                self.delete(rsp_key, rsp_namespace)
                LOGGER.debug("Successfully deleted blob with key: %s", rsp_key)
            except Blob2OverrideError:
                LOGGER.warning("Blob delete skipped due to Blob2OverrideError.")
            except HpIloChifPacketExchangeError:
                LOGGER.warning("Packet exchange error during delete operation.")
            except Exception as excp:
                LOGGER.warning("Error deleting blob: %s", excp)
                raise excp

        return tmpresponse

    def get_security_state(self):
        """Get information for the current security state"""
        LOGGER.debug("Fetching security state from iLO.")

        lib = self.gethprestchifhandle()
        lib.get_security_state.argtypes = []
        lib.get_security_state.restype = POINTER(c_ubyte)

        try:
            # Fetch the security state
            ptr = lib.get_security_state()
            data = ptr[: lib.size_of_securityStateRequest()]
            data = bytearray(data)

            # Send and receive raw data
            resp = self._send_receive_raw(data)

            # Extract the error code from the response
            errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]

            # Check for errors and raise exceptions if needed
            if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
                LOGGER.error("Error occurred with code: %d", errorcode)
                raise HpIloError(errorcode)

            # Attempt to retrieve the security state from the response
            try:
                securitystate = struct.unpack("<c", bytes(resp[72]))[0]
                LOGGER.debug("Security state extracted as character: %s", securitystate)
            except Exception:
                # Fallback for non-character extraction (may be an integer)
                securitystate = int(resp[72])
                LOGGER.debug("Failed to extract character. Security state interpreted as integer: %d", securitystate)

            self.unloadchifhandle(lib)

            LOGGER.debug("Returning security state: %s", securitystate)
            return securitystate

        except Exception as e:
            LOGGER.error("Error while fetching security state: %s", str(e))
            raise

    def mount_blackbox(self):
        """Operation to mount the blackbox partition"""
        lib = self.gethprestchifhandle()
        lib.blackbox_media_mount.argtypes = []
        lib.blackbox_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.blackbox_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def absaroka_media_mount(self):
        """Operation to mount the absaroka repo partition"""
        lib = self.gethprestchifhandle()
        lib.absaroka_media_mount.argtypes = []
        lib.absaroka_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.absaroka_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def gaius_media_mount(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.gaius_media_mount.argtypes = []
        lib.gaius_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.gaius_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def vid_media_mount(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.vid_media_mount.argtypes = []
        lib.vid_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.vid_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def mountflat(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.flat_media_mount.argtypes = []
        lib.flat_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.flat_media_mount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.media_unmount.argtypes = []
        lib.media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def bb_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.bb_media_unmount.argtypes = []
        lib.bb_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.bb_media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def vid_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.vid_media_unmount.argtypes = []
        lib.vid_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.vid_media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def gaius_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.gaius_media_unmount.argtypes = []
        lib.gaius_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.gaius_media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def absr_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.absaroka_media_unmount.argtypes = []
        lib.absaroka_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.absaroka_media_unmount()
        data = ptr[: lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def _send_receive_raw(self, indata):
        """Send and receive raw function for blob operations

        :param indata: The data to be sent to blob operation.
        :type indata: str.

        """
        LOGGER.info("Starting _send_receive_raw operation.")
        excp = None

        for attempt in range(1, 4):  # 3 attempts
            try:
                LOGGER.debug(f"Attempt {attempt}: Sending data to iLO. Data size: {len(indata)} bytes")
                resp = self.channel.send_receive_raw(indata, 10)

                # Log response details
                errorcode = None
                if len(resp) >= 12:
                    errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
                    LOGGER.debug(
                        f"Response details - Size: {len(resp)} bytes, Error Code: {errorcode}, "
                        f"Headers: {resp[:8].hex()}"
                    )
                LOGGER.info("Data successfully sent and received on attempt %d.", attempt)
                return resp
            except Exception as exp:
                LOGGER.warning("Attempt %d failed. Error: %s", attempt, str(exp))
                self.channel.close()
                LOGGER.info("Reinitializing communication channel with iLO.")
                lib = self.gethprestchifhandle()
                self.channel = HpIlo(dll=lib)
                excp = exp  # Store last exception for final raise

        LOGGER.error("All attempts to send/receive raw data have failed.")
        if excp:
            raise excp  # Raise only after all retries fail

    def cert_login(self, cert_file, priv_key, key_pass):
        lib = self.gethprestchifhandle()
        lib.login_cert.argtypes = [c_void_p, c_char_p, c_char_p, c_char_p]
        lib.login_cert.restype = c_char_p

        token = lib.login_cert(self.channel.fhandle, cert_file, priv_key, key_pass)
        return token

    @staticmethod
    def gethprestchifhandle():
        """Multi-platform handle for Chif hprest library"""

        if platform.system() == "Darwin":
            raise ChifDllMissingError()

        LOGGER.debug("Retrieving Chif library handle.")
        excp = None
        libhandle = None
        arch = platform.machine().lower()
        is_windows = os.name == "nt"

        def load_with_flags(path):
            """Load library using RTLD_LAZY | RTLD_GLOBAL to match dlopen behavior."""
            return ctypes.CDLL(path, mode=RTLD_LAZY | RTLD_GLOBAL)

        # 1. Try fixed path on Linux
        if not is_windows:
            libpath = "/opt/ilorest/lib64/libilorestchif.so"
            if os.path.isfile(libpath):
                try:
                    libhandle = load_with_flags(libpath)
                except Exception as exp:
                    excp = exp

        # 2. Try current directory for known filenames
        if not libhandle:
            libnames = ["ilorest_chif.dll"] if is_windows else ["ilorest_chif_dev.so", "ilorest_chif.so"]
            for libname in libnames:
                try:
                    libpath = BlobStore2.checkincurrdirectory(libname)
                    libhandle = load_with_flags(libpath)
                    if libhandle:
                        break
                except Exception as exp:
                    excp = exp

        # 3. Try locating from installed site-packages
        if not libhandle:
            try:
                site_packages = [sysconfig.get_paths()["purelib"]]
            except Exception as exp:
                LOGGER.debug("Failed to get site-packages path: %s", str(exp))
                site_packages = []

            for package in site_packages:
                try:
                    # Base chif library path
                    chif_dir = os.path.join(package, "ilorest", "chiflibrary")

                    if is_windows:
                        chif_lib_name = "ilorest_chif.dll"
                        libpath = os.path.join(chif_dir, chif_lib_name)
                    else:
                        chif_lib_name = "ilorest_chif.so"
                        libpath = os.path.join(chif_dir, chif_lib_name)

                        # Try arch-specific subfolder as fallback
                        if arch in ("aarch64", "arm64"):
                            libpath = os.path.join(chif_dir, "arm", chif_lib_name)

                    if os.path.isfile(libpath):
                        libhandle = load_with_flags(libpath)
                        if libhandle:
                            break
                except Exception as exp:
                    excp = exp

        # Final result
        if libhandle:
            LOGGER.debug("Successfully loaded Chif library.")
            BlobStore2.setglobalhprestchifrandnumber(libhandle)
            return libhandle
        else:
            LOGGER.debug("Failed to load Chif library: %s", str(excp))
            raise ChifDllMissingError(excp)

    @staticmethod
    def setglobalhprestchifrandnumber(libbhndl):
        """Set the random number for the chif handle
        :param libbhndl: The library handle provided by loading the chif library.
        :type libbhndl: library handle.
        """
        rndval = random.randint(1, 65535)
        libbhndl.updaterandval.argtypes = [c_ushort]
        libbhndl.updaterandval(rndval)

    @staticmethod
    def initializecreds(username=None, password=None, log_dir=None):
        """
        Initialize Chif and handle high-security credentials.

        :param username: The username to login.
        :type username: str
        :param password: The password to login.
        :type password: str
        :param log_dir: The directory where logs will be stored.
        :type log_dir: str
        :return: True if successful, False otherwise.
        :rtype: bool
        :raises Blob2SecurityError: If security check fails.
        :raises HpIloInitialError: If credential verification fails.
        """

        LOGGER.info("Initializing Chif credentials with security settings.")
        LOGGER.debug("Username provided: %s, Log directory: %s", username, log_dir)

        try:
            dll = BlobStore2.gethprestchifhandle()

            # Enable debug output if LOGGER level is DEBUG
            if LOGGER.isEnabledFor(logging.DEBUG) and log_dir:
                logdir_c = create_string_buffer(log_dir.encode("utf-8"))
                LOGGER.debug("Enabling debug output to directory: %s", log_dir)
                dll.enabledebugoutput(logdir_c)

            # Initialize Chif
            LOGGER.debug("Calling ChifInitialize()")
            dll.ChifInitialize(None)

            # If username is provided, proceed with security authentication
            if username:
                if not password:
                    LOGGER.warning("Password is missing while username is provided.")
                    return False  # Invalid credentials

                # Check security level requirement
                LOGGER.debug("Checking iLO security requirements")
                security_level = dll.ChifIsSecurityRequired()
                LOGGER.debug(f"Security level check returned: {security_level}")
                if security_level > 0:
                    LOGGER.info("High security mode detected. Authenticating credentials.")
                    LOGGER.debug(f"Security requirements: Username={username}, High Security Mode=True")

                    dll.initiate_credentials.argtypes = [c_char_p, c_char_p]
                    dll.initiate_credentials.restype = POINTER(c_ubyte)

                    usernew = create_string_buffer(username.encode("utf-8"))
                    passnew = create_string_buffer(password.encode("utf-8"))

                    LOGGER.debug("Initiating credential verification.")
                    dll.initiate_credentials(usernew, passnew)

                    credreturn = dll.ChifVerifyCredentials()
                    if credreturn == BlobReturnCodes.SUCCESS:
                        LOGGER.info("Credentials verified successfully.")
                    elif credreturn == hpiloreturncodes.CHIFERR_AccessDenied:
                        LOGGER.error("Access Denied: Invalid credentials.")
                        raise Blob2SecurityError()
                    else:
                        LOGGER.error("Error %s occurred while trying to open a channel to iLO.", credreturn)
                        raise HpIloInitialError(f"Error {credreturn} occurred while trying to open a channel to iLO.")
                else:
                    LOGGER.info("Security not required. Disabling security.")
                    dll.ChifDisableSecurity()
            else:
                LOGGER.debug("No username provided. Checking security requirement.")
                if dll.ChifIsSecurityRequired() > 0:
                    LOGGER.warning("High security mode detected but no credentials provided.")
                    return False
                else:
                    LOGGER.info("Security not required. Disabling security.")
                    dll.ChifDisableSecurity()

            return True

        except Exception as e:
            LOGGER.exception("Exception occurred during Chif initialization: %s", str(e))
            raise

        finally:
            LOGGER.debug("Unloading Chif handle.")
            BlobStore2.unloadchifhandle(dll)

    @staticmethod
    def checkincurrdirectory(libname):
        """Check if the library is present in current directory.
        :param libname: The name of the library to search for.
        :type libname: str."""
        libpath = libname

        if os.path.isfile(os.path.join(os.path.split(sys.executable)[0], libpath)):
            libpath = os.path.join(os.path.split(sys.executable)[0], libpath)
        elif os.path.isfile(os.path.join(os.getcwd(), libpath)):
            libpath = os.path.join(os.getcwd(), libpath)
        elif os.environ.get("LD_LIBRARY_PATH"):
            paths = os.getenv("LD_LIBRARY_PATH", libpath).split(";")
            libpath = [os.path.join(pat, libname) for pat in paths if os.path.isfile(os.path.join(pat, libname))]
            libpath = libpath[0] if libpath else libname

        return libpath

    @staticmethod
    def unloadchifhandle(lib):
        """
        Release a handle on the Chif iLOrest library.

        :param lib: The library handle provided by loading the Chif library.
        :type lib: library handle
        """
        if lib is None:
            LOGGER.warning("unloadchifhandle called with None library handle. Nothing to release.")
            return

        try:
            libhandle = lib._handle
            LOGGER.info("Releasing Chif library handle: %s", libhandle)

            if os.name == "nt":
                LOGGER.debug("Using Windows FreeLibrary to release handle.")
                windll.kernel32.FreeLibrary(None, handle=libhandle)
            else:
                LOGGER.debug("Using dlclose to release handle on Linux/Unix.")
                dlclose(libhandle)

            LOGGER.info("Successfully released Chif library handle.")

        except Exception as e:
            LOGGER.exception("Error while unloading Chif library handle: %s", str(e))



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of viewing the firmware inventory on HPE iLO systems
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def get_inventory_uri(_redfishobj, select):

    update_service_uri = None
    inventory_uri = None

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        update_service_uri = _redfishobj.root.obj['UpdateService']['@odata.id']
    else:
        #Use Resource directory to find the relevant URI
        for instance in resource_instances:
            if '#UpdateService.' in instance['@odata.type']:
                update_service_uri = instance['@odata.id']

    if update_service_uri:
        update_service_resp = _redfishobj.get(update_service_uri)
        if "software" in select.lower():
            inventory_uri = update_service_resp.obj['SoftwareInventory']['@odata.id']
        elif "firmware" in select.lower():
            inventory_uri = update_service_resp.obj['FirmwareInventory']['@odata.id']
        else:
            raise Exception("Invalid selection provided: Please select 'software' or 'firmware' " \
                            "to obtain the relevant invetory data.")
        sys.stdout.write("Printing data in invetory: %s\n" % inventory_uri)
        get_inventory(_redfishobj, inventory_uri)

def get_inventory(_redfishobj, inventory_uri):

    _members = _redfishobj.get(inventory_uri).obj['Members']
    if not _members:
        sys.stderr.write("\tInventory empty.\n")
    else:
        for inventory_item in _members:
            _resp = _redfishobj.get(inventory_item['@odata.id'])
            sys.stdout.write("Printing contents of inventory item, \'%s\':\'%s\'\n" % \
                                            (_resp.dict.get('Name'), _resp.dict.get('Description')))
            print(json.dumps(_resp.dict, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    SELECT = "firmware" #provide either 'software' or 'firmware' for inventory selection
    DISABLE_RESOURCE_DIR = True

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    get_inventory_uri(REDFISHOBJ, SELECT)
    REDFISHOBJ.logout()



###
# Copyright 2025 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Centralized security masking framework for iLORest"""

import json
import re
import copy
from typing import Dict, List, Any, Union, Optional


class SecurityMasker:
    """Centralized framework for masking sensitive information in logs and outputs."""

    # Default mask character
    MASK_VALUE = "******"

    # Number of characters to show for partial masking
    PARTIAL_MASK_SHOW_CHARS = 4

    # Sensitive command-line argument patterns
    SENSITIVE_CLI_KEYWORDS = {
        "--password",
        "--pass",
        "-p",
        "--pwd",
        "--token",
        "--auth-token",
        "--session-token",
        "--key",
        "--api-key",
        "--secret",
        "--credential",
        "--cert-key",
        "--private-key",
        "--passphrase",
        "--sessionkey",
        "--auth",
        "--authorization",
        "--filepass",
        "--backup-password",
        "--biospassword",
    }

    # Fields requiring COMPLETE masking (no characters visible)
    COMPLETE_MASK_FIELDS = {
        "password",
        "oldpassword",
        "newpassword",
        "currentpassword",
        "passwd",
        "pwd",
        "pass",
        "passphrase",
        "secret",
        "client_secret",
        "private_key",
        "privatekey",
        "cert_key",
        "certkey",
        "pin",
        "otp",
        "totp",
        "mfa_token",
        "csrf_token",
        "csrftoken",
    }

    # Fields allowing PARTIAL masking (last 4-6 chars visible for debugging)
    PARTIAL_MASK_FIELDS = {
        "token",
        "auth_token",
        "authtoken",
        "session_token",
        "sessiontoken",
        "access_token",
        "accesstoken",
        "refresh_token",
        "refreshtoken",
        "bearer_token",
        "api_key",
        "apikey",
        "sessionkey",
        "session_key",
        "credential",
        "credentials",
        "appid",
        "app_id",
        "applicationid",
        "application_id",
    }

    # Headers requiring COMPLETE masking
    COMPLETE_MASK_HEADERS = {
        "authorization",  # Often contains Basic auth with password
        "www-authenticate",
        "proxy-authorization",
        "cookie",
        "set-cookie",  # May contain sensitive session data
    }

    # Headers allowing PARTIAL masking (for debugging/correlation)
    PARTIAL_MASK_HEADERS = {
        "x-auth-token",
        "x-session-token",
        "x-api-key",
        "api-key",
        "bearer",  # When separate from Authorization header
        "x-session-id",
        "session-id",
    }

    # All sensitive headers (union of complete and partial)
    SENSITIVE_HEADERS = (
        COMPLETE_MASK_HEADERS
        | PARTIAL_MASK_HEADERS
        | {"authentication", "basic", "digest", "oauth", "jwt", "x-csrf-token", "x-request-id"}
    )

    # All sensitive body fields (union of complete and partial)
    SENSITIVE_BODY_FIELDS = COMPLETE_MASK_FIELDS | PARTIAL_MASK_FIELDS | {"auth", "authorization"}

    # URL parameter patterns that might contain sensitive data
    SENSITIVE_URL_PARAMS = {"password", "token", "key", "secret", "auth", "session", "credential", "pass", "pwd"}

    @classmethod
    def mask_command_arguments(cls, args: List[str]) -> List[str]:
        """
        Mask sensitive information in command-line arguments.

        Args:
            args: List of command-line arguments

        Returns:
            List of arguments with sensitive values masked

        Example:
            >>> SecurityMasker.mask_command_arguments(['--user', 'admin', '--password', 'MyP@ss'])
            ['--user', 'admin', '--password', '******']
        """
        if not args:
            return args

        masked_args = []
        i = 0

        while i < len(args):
            arg = args[i]

            # Check for inline sensitive arguments (--password=value)
            if cls._has_inline_sensitive_value(arg):
                masked_args.append(cls._mask_inline_argument(arg))
                i += 1
                continue

            # Check if current arg is a sensitive flag
            if cls._is_sensitive_cli_flag(arg):
                masked_args.append(arg)  # Keep the flag
                # Mask the next argument if it exists and isn't another flag
                if i + 1 < len(args) and not args[i + 1].startswith("-"):
                    i += 1  # Move to the next argument
                    masked_args.append(cls.MASK_VALUE)  # Mask the value
                i += 1
                continue

            # Keep non-sensitive arguments as-is
            masked_args.append(arg)
            i += 1

        return masked_args

    @classmethod
    def mask_http_headers(
        cls, headers: Union[Dict[str, str], List[tuple], Any]
    ) -> Union[Dict[str, str], List[tuple], str]:
        """
        Mask sensitive information in HTTP headers.

        Args:
            headers: HTTP headers in various formats (dict, list of tuples, or other)

        Returns:
            Headers with sensitive values masked

        Examples:
            # dict format
            >>> SecurityMasker.mask_http_headers({'Authorization': 'Basic abc123', 'Accept': 'application/json'})
            {'Authorization': '******', 'Accept': 'application/json'}

            # list-of-tuples format
            >>> SecurityMasker.mask_http_headers([('Authorization', 'Bearer abcd'), ('Host', 'example')])
            [('Authorization', '******'), ('Host', 'example')]
        """
        if not headers:
            return headers

        try:
            # Handle dictionary format
            if isinstance(headers, dict):
                return cls._mask_headers_dict(headers)

            # Handle list of tuples format
            elif isinstance(headers, (list, tuple)):
                return cls._mask_headers_list(headers)

            # Handle string representation of headers
            elif isinstance(headers, str):
                return cls._mask_headers_string(headers)

            # Handle other object types (e.g., HTTPMessage)
            else:
                return cls._mask_headers_object(headers)

        except Exception:
            # If masking fails, return a safe placeholder
            return "[HEADERS MASKED DUE TO PROCESSING ERROR]"

    @classmethod
    def mask_simple_body(cls, body: Union[str, bytes, Dict[str, Any]]) -> Union[str, bytes, Dict[str, Any]]:
        """
        Mask sensitive information in simple HTTP request/response bodies.

        This method handles basic body masking for simple payloads.
        For complex bodies with advanced JSON parsing and error handling,
        use mask_complex_body() instead.

        Args:
            body: Request/response body in various formats

        Returns:
            Body with sensitive values masked

        Examples:
            # JSON string
            >>> SecurityMasker.mask_simple_body('{"password":"secret"}')
            '{"password":"******"}'

            # dict
            >>> SecurityMasker.mask_simple_body({"token": "abcd"})
            {"token": "******"}
        """
        if not body:
            return body

        try:
            # Handle string body (likely JSON)
            if isinstance(body, str):
                return cls._mask_string_body(body)

            # Handle bytes body
            elif isinstance(body, bytes):
                return cls._mask_bytes_body(body)

            # Handle dictionary body
            elif isinstance(body, dict):
                return cls._mask_dict_body(body)

            # Handle other types
            else:
                return body

        except Exception:
            # If masking fails, return original or safe placeholder
            if isinstance(body, (str, bytes)):
                return "[BODY MASKED DUE TO PROCESSING ERROR]"
            return body

    @classmethod
    def mask_url_parameters(cls, url: str) -> str:
        """
        Mask sensitive information in URL parameters.

        Args:
            url: URL string that may contain parameters

        Returns:
            URL with sensitive parameter values masked

        Example:
            >>> SecurityMasker.mask_url_parameters('https://api.example.com/login?user=admin&password=1234')
            'https://api.example.com/login?user=admin&password=******'
        """
        if not url or "?" not in url:
            return url

        try:
            base_url, params = url.split("?", 1)
            param_pairs = params.split("&")
            masked_pairs = []

            for pair in param_pairs:
                if "=" in pair:
                    key, value = pair.split("=", 1)
                    if any(sensitive in key.lower() for sensitive in cls.SENSITIVE_URL_PARAMS):
                        masked_pairs.append(f"{key}={cls.MASK_VALUE}")
                    else:
                        masked_pairs.append(pair)
                else:
                    masked_pairs.append(pair)

            return f"{base_url}?{'&'.join(masked_pairs)}"

        except Exception:
            return url

    @classmethod
    def mask_complex_body(cls, body: Union[str, bytes, Dict[str, Any]]) -> Union[str, bytes]:
        """
        Mask sensitive information in complex HTTP request/response bodies for logging.

        This is a specialized method for complex body masking that handles:
        - Advanced JSON parsing with error recovery
        - Automatic encoding/decoding (str ↔ bytes)
        - Type preservation (returns same type as input)
        - Graceful error handling to ensure logging never fails

        Use this method for complex request/response bodies. For simple bodies,
        mask_simple_body() may be sufficient.

        Args:
            body: The request/response body in various formats (str, bytes, dict)

        Returns:
            Masked body as string or bytes (matching input type)

        Example:
            >>> SecurityMasker.mask_complex_body(b'{"users":[{"name":"u","password":"p"}]}')
            b'{"users": [{"name": "u", "password": "******"}]}'
        """
        if not body:
            return body

        original_type = type(body)

        try:
            # Convert to string if bytes
            if isinstance(body, bytes):
                try:
                    body = body.decode("utf-8")
                except UnicodeDecodeError:
                    # If decode fails, return original bytes
                    return body

            # If already a dict, mask and return as JSON string
            if isinstance(body, dict):
                masked_dict = cls._mask_dict_body(body)
                return json.dumps(masked_dict)

            # If string, check if it's JSON
            if isinstance(body, str):
                body = body.strip()
                if body.startswith(("{", "[")):
                    try:
                        # Parse JSON and mask
                        body_json = json.loads(body)
                        masked_json = cls._mask_dict_body(body_json)
                        masked_string = json.dumps(masked_json)

                        # Return in original type
                        if original_type == bytes:
                            return masked_string.encode("utf-8")
                        return masked_string
                    except json.JSONDecodeError:
                        # Not valid JSON, return as-is
                        pass

            # Return body as-is if not JSON or other type
            if original_type == bytes and isinstance(body, str):
                return body.encode("utf-8")
            return body

        except Exception:
            # If anything fails, return original to ensure logging continues
            # Converting to string if needed for logging
            if isinstance(body, bytes):
                try:
                    return body.decode("utf-8", errors="replace")
                except:
                    return "[BODY - BINARY DATA]"
            return body if body else ""

    @classmethod
    def mask_user_input(cls, user_input: str, command_context: Optional[str] = None) -> str:
        """
        Mask sensitive information in user input based on context.

        Args:
            user_input: Raw user input string
            command_context: Optional context about what command is being executed

        Returns:
            User input with sensitive parts masked

        Example:
            >>> SecurityMasker.mask_user_input('--login admin --password MyP@ss')
            '--login admin --password ******'
        """
        if not user_input:
            return user_input

        # Split input into tokens for analysis
        tokens = user_input.split()

        # Apply command argument masking
        masked_tokens = cls.mask_command_arguments(tokens)

        return " ".join(masked_tokens)

    # Private helper methods
    @classmethod
    def _is_sensitive_cli_flag(cls, arg: str) -> bool:
        """Check if an argument is a sensitive command-line flag."""
        return arg.lower() in {flag.lower() for flag in cls.SENSITIVE_CLI_KEYWORDS}

    @classmethod
    def _has_inline_sensitive_value(cls, arg: str) -> bool:
        """Check if an argument contains inline sensitive value (flag=value)."""
        if "=" not in arg:
            return False
        flag = arg.split("=")[0]
        return cls._is_sensitive_cli_flag(flag)

    @classmethod
    def _mask_inline_argument(cls, arg: str) -> str:
        """Mask the value part of an inline argument."""
        return re.sub(r"=(.*)", f"={cls.MASK_VALUE}", arg)

    @classmethod
    def _mask_headers_dict(cls, headers: Dict[str, str]) -> Dict[str, str]:
        """Mask sensitive headers in dictionary format."""
        masked = {}
        for key, value in headers.items():
            if cls._is_sensitive_header(key):
                # Determine if complete or partial masking
                masked[key] = cls._mask_value(value, key, is_header=True)
            else:
                masked[key] = value
        return masked

    @classmethod
    def _mask_headers_list(cls, headers: List[tuple]) -> List[tuple]:
        """Mask sensitive headers in list of tuples format."""
        masked = []
        for item in headers:
            if isinstance(item, (list, tuple)) and len(item) >= 2:
                key, value = item[0], item[1]
                if cls._is_sensitive_header(key):
                    masked_value = cls._mask_value(value, key, is_header=True)
                    masked.append((key, masked_value))
                else:
                    masked.append(item)
            else:
                masked.append(item)
        return masked

    @classmethod
    def _mask_headers_string(cls, headers: str) -> str:
        """Mask sensitive headers in string format."""
        lines = headers.split("\n")
        masked_lines = []

        for line in lines:
            if ":" in line:
                key, value = line.split(":", 1)
                if cls._is_sensitive_header(key.strip()):
                    # Preserve leading whitespace (indentation) from original key
                    leading_space = key[: len(key) - len(key.lstrip())]
                    masked_value = cls._mask_value(value.strip(), key.strip(), is_header=True)
                    masked_lines.append(f"{leading_space}{key.strip()}: {masked_value}")
                else:
                    masked_lines.append(line)
            else:
                masked_lines.append(line)

        return "\n".join(masked_lines)

    @classmethod
    def _mask_headers_object(cls, headers: Any) -> str:
        """Mask sensitive headers in object format (e.g., HTTPMessage)."""
        try:
            # Try to convert to dictionary first
            if hasattr(headers, "items"):
                header_dict = dict(headers.items())
                masked_dict = cls._mask_headers_dict(header_dict)
                return str(masked_dict)
            elif hasattr(headers, "__dict__"):
                return cls._mask_headers_dict(headers.__dict__)
            else:
                return "[HEADERS MASKED - UNKNOWN FORMAT]"
        except Exception:
            return "[HEADERS MASKED DUE TO PROCESSING ERROR]"

    @classmethod
    def _is_sensitive_header(cls, header_name: str) -> bool:
        """Check if a header name is considered sensitive."""
        header_lower = header_name.lower().strip()
        return any(sensitive in header_lower for sensitive in cls.SENSITIVE_HEADERS)

    @classmethod
    def _mask_string_body(cls, body: str) -> str:
        """Mask sensitive information in string body."""
        try:
            # Try to parse as JSON
            if body.strip().startswith(("{", "[")):
                json_body = json.loads(body)
                masked_body = cls._mask_dict_body(json_body)
                return json.dumps(masked_body)
            else:
                # Handle form data or other string formats
                return cls._mask_form_data(body)
        except json.JSONDecodeError:
            # Not JSON, treat as form data or plain text
            return cls._mask_form_data(body)

    @classmethod
    def _mask_bytes_body(cls, body: bytes) -> bytes:
        """Mask sensitive information in bytes body."""
        try:
            # Convert to string, mask, then back to bytes
            string_body = body.decode("utf-8")
            masked_string = cls._mask_string_body(string_body)
            return masked_string.encode("utf-8")
        except UnicodeDecodeError:
            # If can't decode, return as-is (likely binary data)
            return body

    @classmethod
    def _mask_dict_body(cls, body: Dict[str, Any]) -> Dict[str, Any]:
        """Mask sensitive information in dictionary body."""
        masked = copy.deepcopy(body)

        def mask_recursive(obj):
            if isinstance(obj, dict):
                for key, value in obj.items():
                    if cls._is_sensitive_body_field(key):
                        # Apply appropriate masking based on field type
                        obj[key] = cls._mask_value(str(value) if value else "", key, is_header=False)
                    else:
                        mask_recursive(value)
            elif isinstance(obj, list):
                for item in obj:
                    mask_recursive(item)

        mask_recursive(masked)
        return masked

    @classmethod
    def _mask_form_data(cls, data: str) -> str:
        """Mask sensitive information in form data."""
        # Handle URL-encoded form data
        if "&" in data and "=" in data:
            pairs = data.split("&")
            masked_pairs = []
            for pair in pairs:
                if "=" in pair:
                    key, value = pair.split("=", 1)
                    if cls._is_sensitive_body_field(key):
                        masked_pairs.append(f"{key}={cls.MASK_VALUE}")
                    else:
                        masked_pairs.append(pair)
                else:
                    masked_pairs.append(pair)
            return "&".join(masked_pairs)

        # For other string formats, return as-is
        return data

    @classmethod
    def _is_sensitive_body_field(cls, field_name: str) -> bool:
        """Check if a body field name is considered sensitive."""
        field_lower = field_name.lower().strip()
        return field_lower in cls.SENSITIVE_BODY_FIELDS

    @classmethod
    def _mask_value(cls, value: str, field_name: str, is_header: bool = False) -> str:
        """
        Mask a sensitive value with either complete or partial masking.

        Args:
            value: The value to mask
            field_name: The field/header name (for determining mask type)
            is_header: Whether this is a header (vs body field)

        Returns:
            Masked value (either completely or partially)
        """
        if not value:
            return cls.MASK_VALUE

        field_lower = field_name.lower().strip()

        # Determine if we should use complete or partial masking
        use_complete_mask = False

        if is_header:
            # Check if header requires complete masking
            use_complete_mask = field_lower in {h.lower() for h in cls.COMPLETE_MASK_HEADERS}
            # Special case: Authorization header with "Basic" is password-based
            if field_lower == "authorization" and value.strip().lower().startswith("basic"):
                use_complete_mask = True
        else:
            # Check if body field requires complete masking
            use_complete_mask = field_lower in {f.lower() for f in cls.COMPLETE_MASK_FIELDS}

        # Apply appropriate masking
        if use_complete_mask:
            return cls.MASK_VALUE
        else:
            # Partial masking: show last N characters
            return cls._partial_mask(value)

    @classmethod
    def _partial_mask(cls, value: str, show_chars: int = None) -> str:
        """
        Apply partial masking to a value, showing only the last N characters.

        Args:
            value: The value to mask
            show_chars: Number of characters to show (default: PARTIAL_MASK_SHOW_CHARS)

        Returns:
            Partially masked value like "******a0fb"
        """
        if show_chars is None:
            show_chars = cls.PARTIAL_MASK_SHOW_CHARS

        if not value or len(value) <= show_chars:
            # If value is too short, completely mask it
            return cls.MASK_VALUE

        # Show last N characters
        visible_part = value[-show_chars:]
        return f"{cls.MASK_VALUE}{visible_part}"


# Convenience functions for backward compatibility and easy usage
def mask_passwords(args: List[str]) -> List[str]:
    """Backward compatible function for masking passwords in command arguments."""
    return SecurityMasker.mask_command_arguments(args)


def mask_http_request_data(method: str, path: str, headers: Any, body: Any) -> tuple:
    """
    Mask all sensitive data in an HTTP request for logging.

    Args:
        method: HTTP method
        path: Request path/URL
        headers: Request headers
        body: Request body

    Returns:
        Tuple of (method, masked_path, masked_headers, masked_body)
    """
    masked_path = SecurityMasker.mask_url_parameters(path)
    masked_headers = SecurityMasker.mask_http_headers(headers)
    masked_body = SecurityMasker.mask_simple_body(body)

    return method, masked_path, masked_headers, masked_body


def mask_user_input(user_input: str, command_context: Optional[str] = None) -> str:
    """Convenience function for masking user input."""
    return SecurityMasker.mask_user_input(user_input, command_context)
